(* Delphi program
   Display the Mandelbrot set
   ==========================

    Dr. J. Rathlev, D-24222 Schwentinental (kontakt(a)rathlev-home.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 2.0 : February 2021
   *)

unit MandelMain;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes,
  Vcl.Graphics, Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.ExtCtrls, Vcl.StdCtrls,
  System.Contnrs, Vcl.Buttons, NumberEd;

const
  IterMax = 256;
  defSize = 500;

type
  TUserPal = record
    lpal  : TLogPalette;
    dummy : Array[0..255] of TPaletteEntry;
    end;

  TRGB = record
    b,g,r : byte;
    end;
  PRGB = ^TRGB;

  TxPal = Array [0..255] of TRGB;

  TFarbPal = record
    FarbAnz   : integer;
    FarbTab   : TxPal;
    end;

  TEditMode = (edNoth,edNew,edSize,edMove);

  TField = class(TObject)
    X0,Y0,DX : extended;
    bm       : TBitMap;
    Inverse  : boolean;
    constructor Create (x,y,d : extended; Inv : boolean; b : TBitMap);
    destructor Destroy; override;
    end;

  TfrmMandel = class(TForm)
    paSettings: TPanel;
    GroupBox1: TGroupBox;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    Label6: TLabel;
    edtX0: TFloatEdit;
    edtY0: TFloatEdit;
    edtDX: TFloatEdit;
    btnEnde: TBitBtn;
    btnStart: TBitBtn;
    btnStop: TBitBtn;
    paImage: TPanel;
    Label5: TLabel;
    btnReset: TBitBtn;
    PaintBox: TPaintBox;
    btnPrev: TSpeedButton;
    btnNext: TSpeedButton;
    btnDelete: TSpeedButton;
    cbInvers: TCheckBox;
    btnClipboard: TBitBtn;
    cbCoord: TCheckBox;
    procedure btnStartClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure btnStopClick(Sender: TObject);
    procedure btnEndeClick(Sender: TObject);
    procedure btnResetClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure PaintBoxPaint(Sender: TObject);
    procedure PaintBoxMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure PaintBoxMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure PaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure btnPrevClick(Sender: TObject);
    procedure btnNextClick(Sender: TObject);
    procedure btnDeleteClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure btnClipboardClick(Sender: TObject);
  private
    { Private-Deklarationen }
    FrameOn,
    Stopped   : boolean;
    BitMap    : array [boolean] of TBitMap;
    FramePos  : TRect;
    EdMode           : TEditMode;
    OldMPos,MPos     : TPoint;
    Fields           : TObjectList;
    SelectedField    : integer;
    procedure UpdateButtons;
    procedure ShowImage;
    procedure MouseToXY (var Pt : TPoint);
    procedure XYToMouse (var Pt : TPoint);
    procedure DrawRect (ARect : TRect);
    procedure ShowFrame;
    function InFrame (Pos : TPoint) : boolean;
    procedure MoveFrame;
    procedure RemoveFrame;
  public
    { Public-Deklarationen }
  end;

var
  frmMandel: TfrmMandel;

implementation

{$R *.dfm}

uses Vcl.Clipbrd, WinUtils;

constructor TField.Create (x,y,d : extended; Inv : boolean; b : TBitMap);
begin
  inherited Create;
  X0:=x; Y0:=y; DX:=d; Inverse:=Inv;
  bm:=TBitMap.Create;
  bm.Assign(b);
  end;

destructor TField.Destroy;
begin
  bm.Free;
  inherited Destroy;
  end;

procedure TfrmMandel.FormCreate(Sender: TObject);
var
  FPal    : TFarbPal;
  lp      : PLogPalette;

  procedure SetPalette(const FarbPal : TFarbpal);
  var
    i : integer;
  begin
    if assigned(lp) then FreeMem(lp);
    GetMem(lp,SizeOf(TLogPalette)+(FarbPal.FarbAnz-1)*SizeOf(TPaletteEntry));
    with lp^ do begin
      palVersion:=$300;
      palNumEntries:=FarbPal.FarbAnz;
      for i:=0 to 255 do with PalpalEntry[i],FarbPal.FarbTab[i] do begin
        peRed:=r; peGreen:=g; peBlue:=b; peFlags:=0;
        end;
      end;
    end;

  procedure AssignColors(Start,Delta : integer; Inverse : boolean; var FarbTab : TxPal);
  var
    h,i      : integer;
  const
    ColSteps = 35;

    function ColorScaleInc (w,ColSteps : integer) : byte;
    begin
      Result:=round(255*(1-sqr((w-ColSteps)/ColSteps)));
      end;

    function ColorScaleDec (w,ColSteps : integer) : byte;
    begin
      Result:=255-round(255*sqr(w/ColSteps));
      end;

  begin
    for i:=0 to 255 do with FarbTab[i] do begin
      h:=round(256*Delta*i/255+Start) mod (7*ColSteps);
      if h<=ColSteps then begin
        g:=0; b:=0;
        r:=ColorScaleInc(h,ColSteps);
        end
      else if h<=2*ColSteps then begin
        r:=255; b:=0;
        g:=ColorScaleInc(h-ColSteps,ColSteps);
        end
      else if h<=3*ColSteps then begin
        g:=255; b:=0;
        r:=ColorScaleDec(h-2*ColSteps,ColSteps);
        end
      else if h<=4*ColSteps then begin
        g:=255; r:=0;
        b:=ColorScaleInc(h-3*ColSteps,ColSteps);
        end
      else if h<=5*ColSteps then begin
        b:=255; r:=0;
        g:=ColorScaleDec(h-4*ColSteps,ColSteps);
        end
      else if h<=6*ColSteps then begin
        b:=255; g:=0;
        r:=ColorScaleInc(h-5*ColSteps,ColSteps);
        end
      else begin
        b:=255; r:=255;
        g:=ColorScaleInc(h-6*ColSteps,ColSteps);
        end;
      if Inverse then begin
        r:=255-r; g:=255-g; b:=255-b;
        end;
      end;
   end;

begin
  lp:=nil;
  with FPal do begin
    FarbAnz:=256;
    AssignColors(0,1,false,FarbTab);
    end;
  SetPalette(FPal);
  BitMap[false]:=TBitMap.Create;
  with BitMap[false] do begin
    PixelFormat:=pf8bit;
    Palette:=CreatePalette(lp);
    Width:=PaintBox.Width; Height:=PaintBox.Height;
    with Canvas do begin
      Brush.Color:=clBlack;
      FillRect(Rect(0,0,Width-1,Height-1));
      end;
    end;
  with FPal do begin
    FarbAnz:=256;
    AssignColors(0,1,true,FarbTab);
    end;
  SetPalette(FPal);
  BitMap[true]:=TBitMap.Create;
  with BitMap[true] do begin
    PixelFormat:=pf8bit;
    Palette:=CreatePalette(lp);
    Width:=PaintBox.Width; Height:=PaintBox.Height;
    with Canvas do begin
      Brush.Color:=clBlack;
      FillRect(Rect(0,0,Width-1,Height-1));
      end;
    end;
  Fields:=TObjectList.Create;
  with Fields do begin
    OwnsObjects:=true;
    end;
  FrameOn:=false;
  SelectedField:=-1;
  btnStop.Enabled:=false;
  with PaintBox.Canvas do begin
    Font.Height:=-18;
    end;
  UpdateButtons;
  end;

procedure TfrmMandel.FormDestroy(Sender: TObject);
begin
  BitMap[false].Free; BitMap[true].Free; Fields.Free;
  end;

procedure TfrmMandel.FormResize(Sender: TObject);
begin
  ClientWidth:=ClientHeight+PaSettings.Width;
  PaintBox.Invalidate;
  end;

procedure TfrmMandel.FormShow(Sender: TObject);
begin
  btnStartClick(Sender);
  end;

procedure TfrmMandel.UpdateButtons;
begin
  with Fields do begin
    btnNext.Enabled:=(Count>1) and (SelectedField<Count-1);
    btnPrev.Enabled:=(Count>1) and (SelectedField>0);
    btnDelete.Enabled:=btnPrev.Enabled;
    end;
  end;

procedure TfrmMandel.ShowImage;
begin
  with Fields.Items[SelectedField] as TField do begin
    edtX0.Value:=X0; edtY0.Value:= Y0; edtDX.Value:=DX;
    cbInvers.Checked:=Inverse;
    BitMap[Inverse].Assign(bm);
    end;
  UpdateButtons;
  PaintBox.Invalidate;
  end;

procedure TfrmMandel.btnPrevClick(Sender: TObject);
begin
  FrameOn:=false;
  dec(SelectedField);
  ShowImage;
  end;

procedure TfrmMandel.btnNextClick(Sender: TObject);
begin
  FrameOn:=false;
  inc(SelectedField);
  ShowImage;
  end;

procedure TfrmMandel.btnClipboardClick(Sender: TObject);
begin
  ClipBoard.Assign(BitMap[cbInvers.Checked]);
  end;

procedure TfrmMandel.btnDeleteClick(Sender: TObject);
begin
//  if MessageDlgPos('Aktuelles Bild lschen?',mtConfirmation,mbOKCancel,0,CursorPos.X,CursorPos.Y,mbOK)=mrOK then begin
  if (SelectedField>0) and ConfirmDialog(CursorPos,'Aktuelles Bild lschen?') then begin
    FrameOn:=false;
    with Fields do begin
      Delete(SelectedField);
      if SelectedField>=Count then SelectedField:=Count-1;
      end;
    ShowImage;
    end;
  end;

procedure TfrmMandel.btnStartClick(Sender: TObject);
var
  xi,yi,x0,y0,
  xz,yz,x,y   : extended;
  xm,ym,i,j,k : integer;
  P           : PByteArray;

  procedure coper(var xz,yz : extended; x,y : extended);
  var
    x1 : extended;
  begin
    x1:=xz*xz-yz*yz+x; yz:=2*xz*yz+y; xz:=x1;
    end;

begin
  Stopped:=false;
  btnStop.Enabled:=true;
  with Bitmap[cbInvers.Checked] do begin
    xm:=Width; ym:=Height;
    end;
  x0:=edtX0.Value; y0:=edtY0.Value;
  xi:=edtDX.Value/xm;
  xz:=PaintBox.Width/xm;
  yz:=PaintBox.Height/ym;
  if FrameOn then with FramePos do begin
    x0:=x0+xi*Left/xz; y0:=y0+xi*(ym-Bottom/yz);
    edtX0.Value:=x0;
    edtY0.Value:=y0;
    edtDX.Value:=xi*abs(Right-Left)/xz;
    xi:=edtDX.Value/xm;
    Left:=0; Right:=0; Top:=0; Bottom:=0;
    FrameOn:=false;
    end;
  yi:=xi; y:=y0;
  with BitMap[cbInvers.Checked].Canvas do begin
    Brush.Color:=clWhite;
    FillRect(Rect(0,0,Width,Height));
    end;
  for i:=1 to ym do begin
    x:=x0;
    P:=BitMap[cbInvers.Checked].ScanLine[ym-i];
    for j:=0 to xm-1 do begin
      xz:=x; yz:=y; k:=-1;
      while k<IterMax do begin
        coper (xz,yz,x,y); inc(k);
        if xz*xz+yz*yz>4 then break;
        end;
      P[j]:=k mod 256;
      if Stopped then Exit;
      x:=x+xi;
      end;
    Application.ProcessMessages;
    with PaintBox do Canvas.StretchDraw(Rect(0,0,Width-1,Height-1),BitMap[cbInvers.Checked]);
    y:=y+yi;
    end;
  if cbCoord.Checked then with BitMap[cbInvers.Checked],Canvas do begin
    Font.Color:=clBlack;
    Brush.Color:=clWhite;
    TextOut(20,Height-25,' X: '+edtX0.Text+' Y: '+edtY0.Text+' D:'+edtDx.Text+' ');
    end;
  PaintBox.Invalidate;
  inc(SelectedField);
  Fields.Insert(SelectedField,TField.Create(edtX0.Value,edtY0.Value,edtDX.Value,
                cbInvers.Checked,BitMap[cbInvers.Checked]));
  btnStop.Enabled:=false;
  UpdateButtons;
  end;

procedure TfrmMandel.btnStopClick(Sender: TObject);
begin
  Stopped:=true;
  end;

procedure TfrmMandel.btnEndeClick(Sender: TObject);
begin
  Stopped:=true;
  Close;
  end;

procedure TfrmMandel.btnResetClick(Sender: TObject);
begin
  edtX0.Value:=-1.5;
  edtY0.Value:=-1.0;
  edtDx.Value:=2.0;
  cbInvers.Checked:=false;
  end;

procedure TfrmMandel.PaintBoxPaint(Sender: TObject);
begin
  with PaintBox,Canvas do begin
    StretchDraw(Rect(0,0,Width-1,Height-1),BitMap[cbInvers.Checked]);
    end;
  if FrameOn then DrawRect(FramePos);
  end;

{-----------------------------------------------------------------------}
(* Bildschirmkoordinaten umrechnen *)
procedure TfrmMandel.MouseToXY (var Pt    : TPoint);
var
  PListe : array [1..1] of TPoint absolute Pt;
begin
  DPToLP (PaintBox.Canvas.Handle,PListe,1);
  end;

procedure TfrmMandel.XYToMouse (var Pt    : TPoint);
var
  PListe : array [1..1] of TPoint absolute Pt;
begin
  LPToDP (PaintBox.Canvas.Handle,PListe,1);
  end;

{-----------------------------------------------------------------------}
procedure TfrmMandel.DrawRect (ARect : TRect);
begin
  with PaintBox.Canvas do begin
    with Pen do begin
      Color:=clWhite;
      Mode:=pmXor;
      Style:=psSolid;
      Width:=1;
      end;
    with Brush do begin
      Style:=bsClear;
      end;
    Rectangle(ARect);
    end;
  end;

function TfrmMandel.InFrame (Pos : TPoint) : boolean;
begin
  MouseToXY(Pos);
  if FrameOn then
    with FramePos do Result:=(Left<=Pos.X) and (Right>=Pos.X)
                     and (Top<=Pos.Y) and (Bottom>=Pos.Y)
  else Result:=false
  end;

procedure TfrmMandel.ShowFrame;
begin
  FrameOn:=true;
  DrawRect (FramePos);
  end;

procedure TfrmMandel.RemoveFrame;
begin
  DrawRect (FramePos);
  FrameOn:=false;
  end;

procedure TfrmMandel.MoveFrame;
var
  Pt      : TPoint;
  dx,dy   : integer;
begin
  with Pt do begin
    x:=MPos.X; y:=MPos.Y;
    end;
  with Pt do begin
    dx:=X; dy:=Y;
    x:=OldMPos.X; y:=OldMPos.Y;
    end;
  with Pt do begin
    dx:=dx-X; dy:=dy-Y;
    end;
  with FramePos do begin
    if dx+Left<0 then dx:=0;
    if dy+Top<0 then dy:=0;
    if dx+Right>=PaintBox.Width then dx:=0;
    if dy+Bottom>=PaintBox.Height then dy:=0;
    Left:=Left+dx; Top:=Top+dy;
    Right:=Right+dx; Bottom:=Bottom+dy;
    DrawRect (FramePos);
    end;
  FrameOn:=true;
  end;

procedure TfrmMandel.PaintBoxMouseDown(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  MPos.X:=X; MPos.Y:=Y;
  if Button=mbLeft then begin
    if InFrame(MPos) then begin  // bewegen
      EdMode:=edMove;
      end
    else begin  // neues Feld
      EdMode:=edNew;
      RemoveFrame;
      with FramePos do begin
        Left:=MPos.X; Top:=MPos.Y;
        Right:=MPos.X; Bottom:=MPos.Y;
        end;
      PaintBox.Invalidate;
      end;
    OldMPos:=MPos;
    end
  else if Button=mbRight then RemoveFrame;
  end;

procedure TfrmMandel.PaintBoxMouseMove(Sender: TObject; Shift: TShiftState;
  X, Y: Integer);
begin
  MPos.X:=X; MPos.Y:=Y;
  if edMode<>edNoth then begin
    if EdMode=edMove then begin
      RemoveFrame;
      MoveFrame;
      OldMPos:=MPos;
      end
    else begin // neu
      RemoveFrame;
      with FramePos do begin
        Left:=OldMPos.X; Top:=OldMPos.Y;
        Right:=MPos.X; Bottom:=Top+Right-Left;
        end;
      ShowFrame;
      end;
    end;
  end;

procedure TfrmMandel.PaintBoxMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  if (EdMode<>edNoth) and (Button=mbLeft) then begin
    if (EdMode=edNew) then begin // edNew
      end
    else if EdMode=edMove then begin
      end;
    RemoveFrame;
    ShowFrame;
    PaintBox.Invalidate;
    EdMode:=edNoth;
    end;
  end;

end.
